import java.io.*;
import java.util.Vector;
import java.lang.Math;

/*
	Version 2
	Changes:
		- change: print dots during conversion
		- fix: multiple attachments
		- change: attachments marked as attachment (not as inline)
		- change: also reads MW 3.61 folders (exprimental)

	Version 2.1
	Changes:
		- bugfix: crash if folder index not ordered

	Version 2.2
	Changes:
		- bugfix: crash if file attachment too long
		- sent-items: "From:", "Subject:" and "Date:" line are written in any case
		- empty line between mails
		- bugfix: sent-items: no "Content-Type:" line even if attachments available

	Version 2.3
	Changes:
		- bugfix: hangs if mail has no body
*/

public class Qdb 
{
    private MyInputStream in;
//    private PrintWriter out;
    private int totalSize, pos;
    private Vector mailIndex = new Vector();
    private static int maxFileSize = 5000000;

    private class MyInputStream extends DataInputStream
    {
        public MyInputStream(InputStream in)
        {
            super(in);
        }
        
        public String readLn() throws Exception
        {
            String result = "";
            char c;
            do
            {
                c = (char)read();
                if (c != 13 && c != 10)
                    result += c;
            }
            while (c != 10);
            return result;
        }
    }
    
    private class Date
    {
        public int year, month, day, hour, minute, second, millisecond, dayOfWeek;
        private int  ONE_SECOND = 1000;
        private int  ONE_MINUTE = 60*ONE_SECOND;
        private int  ONE_HOUR   = 60*ONE_MINUTE;
        private long ONE_DAY    = 24*ONE_HOUR;
        
        public String getWeekDay()
        {
            switch (dayOfWeek)
            {
                case 1: return "Sun";
                case 2: return "Mon";
                case 3: return "Tue";
                case 4: return "Wed";
                case 5: return "Thu";
                case 6: return "Fri";
                case 7: return "Sat";
                default: return "???";
            }
        }
        
        public String getMonth()
        {
            switch (month)
            {
                case 1: return "Jan";
                case 2: return "Feb";
                case 3: return "Mar";
                case 4: return "Apr";
                case 5: return "May";
                case 6: return "Jun";
                case 7: return "Jul";
                case 8: return "Aug";
                case 9: return "Sep";
                case 10: return "Oct";
                case 11: return "Nov";
                case 12: return "Dec";
                default: return "???";
            }
        }
        
        private long floorDivide(long numerator, long denominator) 
        {
            // We do this computation in order to handle
            // a numerator of Long.MIN VALUE correctly
            return (numerator >= 0) ?
                numerator / denominator :
                ((numerator + 1) / denominator) - 1;
        }
        
        private int floorDivide(long numerator, int denominator, int[] remainder) 
        {
            if (numerator >= 0) {
                remainder[0] = (int)(numerator % denominator);
                return (int)(numerator / denominator);
            }
            int quotient = (int)(((numerator + 1) / denominator) - 1);
            remainder[0] = (int)(numerator - (quotient * denominator));
            return quotient;
        }
        
        private long millisToJulianDay(long millis) 
        {
            int EPOCH_JULIAN_DAY   = 2440588; // Jaunary 1, 1970 (Gregorian)
            return EPOCH_JULIAN_DAY + floorDivide(millis, ONE_DAY);
        }
        
        public void setTime(long millis)
        {
            long gregorianCutover = -12219292800000L;
            int JAN_1_1_JULIAN_DAY = 1721426; // January 1, year 1 (Gregorian)
            int LEAP_NUM_DAYS[] = {0,31,60,91,121,152,182,213,244,274,305,335}; // 0-based, for day-in-year
            int NUM_DAYS[] = {0,31,59,90,120,151,181,212,243,273,304,334}; // 0-based, for day-in-year
            int SUNDAY = 1;
            long normalizedGregorianCutover = gregorianCutover;
            boolean isLeap;
            int dayOfYear, rawYear;

            // Compute the year, month, and day of month from the given millis
            if (millis >= normalizedGregorianCutover) 
            {
                // The Gregorian epoch day is zero for Monday January 1, year 1.
                long gregorianEpochDay = millisToJulianDay(millis) - JAN_1_1_JULIAN_DAY;
                // Here we convert from the day number to the multiple radix
                // representation.  We use 400-year, 100-year, and 4-year cycles.
                // For example, the 4-year cycle has 4 years + 1 leap day; giving
                // 1461 == 365*4 + 1 days.
                int[] rem = new int[1];
                int n400 = floorDivide(gregorianEpochDay, 146097, rem); // 400-year cycle length
                int n100 = floorDivide(rem[0], 36524, rem); // 100-year cycle length
                int n4 = floorDivide(rem[0], 1461, rem); // 4-year cycle length
                int n1 = floorDivide(rem[0], 365, rem);
                rawYear = 400*n400 + 100*n100 + 4*n4 + n1;
                dayOfYear = rem[0]; // zero-based day of year
                if (n100 == 4 || n1 == 4) {
                    dayOfYear = 365; // Dec 31 at end of 4- or 400-yr cycle
                } else {
                    ++rawYear;
                }

                isLeap = ((rawYear&0x3) == 0) && // equiv. to (rawYear%4 == 0)
                    (rawYear%100 != 0 || rawYear%400 == 0);
                dayOfWeek = (int)((gregorianEpochDay+1) % 7);

            }
            else 
            {
                // The Julian epoch day (not the same as Julian Day)
                // is zero on Saturday December 30, 0 (Gregorian).
                long julianEpochDay = millisToJulianDay(millis) - (JAN_1_1_JULIAN_DAY - 2);
                rawYear = (int) floorDivide(4*julianEpochDay + 1464, 1461);

                // Compute the Julian calendar day number for January 1, rawYear
                long january1 = 365*(rawYear-1) + floorDivide(rawYear-1, 4);
                dayOfYear = (int)(julianEpochDay - january1); // 0-based

                // Julian leap years occurred historically every 4 years starting
                // with 8 AD.  Before 8 AD the spacing is irregular; every 3 years
                // from 45 BC to 9 BC, and then none until 8 AD.  However, we don't
                // implement this historical detail; instead, we implement the
                // computatinally cleaner proleptic calendar, which assumes
                // consistent 4-year cycles throughout time.
                isLeap = ((rawYear&0x3) == 0); // equiv. to (rawYear%4 == 0)

                dayOfWeek = (int)((julianEpochDay-1) % 7);
            }
        
            // Common Julian/Gregorian calculation
            int correction = 0;
            int march1 = isLeap ? 60 : 59; // zero-based DOY for March 1
            if (dayOfYear >= march1) correction = isLeap ? 1 : 2;
            month = (12 * (dayOfYear + correction) + 6) / 367; // zero-based month
            day = dayOfYear - (isLeap ? LEAP_NUM_DAYS[month] : NUM_DAYS[month]) + 1; // one-based DOM
            dayOfWeek += (dayOfWeek < 0) ? (SUNDAY+7) : SUNDAY;

            year = rawYear;

            month++;
            long fraction = millis % ONE_DAY;
            millisecond = (int)(fraction % 1000);
            fraction /= 1000;
            second = (int)(fraction % 60);
            fraction /= 60;
            minute = (int)(fraction % 60);
            fraction /= 60;
            hour = (int)fraction;
        }
        
        private String twoDigit(int i)
        {
            String result = Integer.toString(i);
            if (result.length() < 2)
                result = "0" + result;
            return result;
        }
        
        private String threeDigit(int i)
        {
            String result = twoDigit(i);
            if (result.length() < 3)
                result = "0" + result;
            return result;
        }
        
        public String toAscTimeInFromField()
        {
            String result = getWeekDay() + " " + getMonth();
            String s = Integer.toString(day);
            if (s.length() == 1)
                s = " " + s;
            result += " " + s + " " + twoDigit(hour) + ":" + twoDigit(minute) + ":" + twoDigit(second) + " " + Integer.toString(year);
            return result;
        }
        
        public String toAscTimeInDateField()
        {
            String result = getWeekDay() + ", " + day + " " + getMonth() + " " + year+ " " + twoDigit(hour) + ":" + twoDigit(minute) + ":" + twoDigit(second) + " GMT";
            return result;
        }
        
        public String toString()
        {
            return year + "-" + twoDigit(month) + "-" + twoDigit(day) + " " + twoDigit(hour) + ":" + twoDigit(minute) + ":" + twoDigit(second) + "." + threeDigit(millisecond) + " (" + getWeekDay() + ")";
        }
    }
    
    private void main(File sourceFile, boolean withAttatchments, int mwVersion)
    {
	PrintWriter out = null;
//Mail m = new Mail();
//System.out.println(encodeFile("G:\\Program Files\\Microsoft Hardware\\Mouse\\Ergo5.gif"));
//System.exit(0);
	try
	{
	    System.out.println("processing: " + sourceFile.getName());
            FileInputStream f = new FileInputStream(sourceFile);
            in = new MyInputStream(f);
            if (in.read() != 'Q' || in.read() != 'D' || in.read() != 'B')
                throw new Exception("Not a QDB file");
            in.read();
            totalSize = readInt();
            for (int i = 0; i < totalSize; i++)
	        in.read();
            if (in.read() != 'Q' || in.read() != 'I' || in.read() != 'X')
                throw new Exception("Missing QIX part");
            in.read();
            totalSize = readInt()-4;
            int num = readInt();
//System.out.println("-------------------------QIX-------------------------------------");
//System.out.println(totalSize + ", " + num);
//System.out.println("--------------------------------------------------------------");
            int i = 0;
	    System.out.print("Part 1: folder index: ");
	    mailIndex = new Vector();
            for (;;)
            {
                int size = 0;
                ++i;
                if (totalSize <= 0)
                    break;
	        MailIndex index = new MailIndex();
                index.qixString = "";
                for(;;)
                {
                    char c = (char)in.read();
                    ++size;
		    if (size > 100)
		        throw new Exception ("synch lost");
                    if (c == 0)
                        break;
                    index.qixString += c;
                }
                index.pos = readInt();
                index.len = readInt();
                index.qixInt2 = readInt();
		int pos = mailIndex.size();
		for (int j = 0; j < mailIndex.size(); j++)
		    if (((MailIndex)mailIndex.get(j)).pos > index.pos)
		    {
			pos = j;
			break;
		    }
		mailIndex.insertElementAt(index, pos);
	    
                size += 12;
                totalSize -= size;
                boolean found = false;
                if (totalSize <= 0)
                    break;
		System.out.print(".");
            }
//for (int j = 0; j < mailIndex.size(); j++)System.out.println(mailIndex.get(j));
	    System.out.println();
	    in.close();
	    if (mailIndex.size() == 0)
		return;
	
	    String targetFileName = sourceFile.getName().substring(0, sourceFile.getName().length()-4) + ".mbox";
	    if (mwVersion == 1)
	    {
		String s = sourceFile.getName();
		s = s.substring(0, s.length()-4);
		targetFileName = sourceFile.getParent() + "." + s + ".mbox";
	    }
//targetFileName = targetFileName.replace('@', '#');
            out = new PrintWriter(new FileOutputStream(targetFileName));
            f = new FileInputStream(sourceFile);
            in = new MyInputStream(f);
            if (in.read() != 'Q' || in.read() != 'D' || in.read() != 'B')
                throw new Exception("Not a QDB file");
            in.read();
            totalSize = readInt();

            pos = 0;
//System.out.println(totalSize);
//System.out.println("--------------------------------------------------------------");
            i = 0;
	    System.out.println("Part 2: mails. Target file: " + targetFileName);
            while (true)
            {
                ++i;
                if (pos >= totalSize)
                    break;
                int oldpos = pos;
		int mailSize = totalSize; //3.61 only
		if (mailIndex.size() >= 1)
		    mailSize = ((MailIndex)mailIndex.get(0)).len;
                Mail mail = readMail(mailSize, mwVersion);
                if (mail == null)
                    break;
                mail.pos = oldpos;
	        for (int j = 0; j < mailIndex.size(); j++)
	        {
		    MailIndex index = (MailIndex)mailIndex.get(j);
		    if (index.pos == mail.pos)
		    {
		        mail.qixInt1 = index.len;
		        mail.qixInt2 = index.qixInt2;
		        mail.qixString = index.qixString;
		        mailIndex.remove(j);
		        break;
		    }
	        }
                if (mail.qixString != null)
                {
//System.out.println(mail);
//System.out.println(i + "--------------------------------------------------------------");
	    	    mail.toMbox(withAttatchments, out);
                }
		System.out.print(".");
            }
	    System.out.println();
if (mailIndex.size() != 0)
{System.out.println();
System.out.println("Error: mailindex not empty: " + mailIndex.size());}
	}
	catch (Exception e)
	{
	    e.printStackTrace();
	}
	if (out != null)
	    try
	    {
	    	out.close();
	    }
            catch (Exception e)
            {
                e.printStackTrace();
            }
    }

    private class MailIndex
    {
	public int pos;
        public int len /* at least on 3.61 */, qixInt2;
        public String qixString;
        public String toString()
	{
	    return qixString + ", " + len + ", " + qixInt2 + ", pos: " + pos;
	}
    }

    private class Mail
    {
        public int unknownInt1, pos;
        public int qixInt1, qixInt2;
        public String qixString = null;
        public Date date;
        public Vector attachments = new Vector();
        public Vector attachmentsFound = new Vector();
        public Vector addLines = new Vector();
        public String subject, from1, from2, from3, to, cc, bcc, account, mime, id, guessedMime;
	public Vector body = new Vector();
        public String unknown1, unknown3, unknown4, unknown5, unknown6;
        public String terminator, boundary;
        
        public String toString()
        {
            String a = "", a2 = "";
            for (int i = 0; i < attachments.size(); i++)
            {
                if (i !=0)
                    a = ", " + a;
                a += attachments.get(i);
            }
            for (int i = 0; i < attachmentsFound.size(); i++)
            {
                if (i !=0)
                    a2 = ", " + a2;
                a2 += attachmentsFound.get(i);
            }

            String b = new String();
	    for (int i = 0; i < body.size(); i++)
		b += (String)body.get(i) + "\\n";
            String c = "";
            for (int i = 0; i < addLines.size(); i++)
                c += "Additional line " + Integer.toString(i+1) + ": " + addLines.get(i) + "\n";
            String u = Integer.toString(unknownInt1) + ", " + unknown1 + ", " + ", " + unknown3 + ", " + unknown4 + ", " + unknown5 + ", " + unknown6;
            String q = "---none---";
            if (qixString != null)
                q = qixInt1 + ", " + qixInt1 + ", " + qixString;
            return "Subject: " + subject + "\nfrom1: " + from1 + "\nfrom2: " + from2 + "\nfrom3: " + from3 + "\nto: " + to + "\ncc: " + cc + "\nbcc: " + bcc + "\naccout: " + account + "\nmime: " + mime + "\nguessed mime: " + guessedMime + "\ndate: " + date + "\nattachments: " + a + "\nexisting attachments: " + a2 + "\nid: " + id +  "\nboundary: " + boundary + "\nunknown: " + u + "\nQIX:" + q + "\n" + c + "\nbody: " + b;
        }
        
        private boolean isLetter(char c)
        {
            return (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z');
        }
        
        public void toMbox(boolean withAttatchments, PrintWriter out)
        {
            String from = from3;
	    boolean fromLine = false, dateLine = false, subjectLine = false, contentTypeLine = false;
            int i = from.indexOf(' ');
            if (i > 0)
                from = from.substring(0, i);
            out.println("From " + from + " " + date.toAscTimeInFromField());
            boolean toPartProcessed = false;
            for (i = 0; i < addLines.size(); i++)
            {
            	String s = (String)addLines.get(i);
            	String t = s.toLowerCase();
            	boolean removeContinuedLine = false;
    	        while (t.startsWith("to: ") || t.startsWith("cc: ") || t.startsWith("bcc: ") || (removeContinuedLine && s.length() > 0 && !isLetter(s.charAt(0))))
    	        {
    	            removeContinuedLine = true;
    	    	    ++i;
    	    	    if (i < addLines.size())
    	    	    {
            	        s = (String)addLines.get(i);
            	        t = s.toLowerCase();
            	    }
            	    else
            	    {
            	    	s = "";
            	    	t = "";
            	        break;
            	    }
    	        }
            	if (s.startsWith("From: "))
		    fromLine = true;
            	if (s.startsWith("Date: "))
		    dateLine = true;
            	if (s.startsWith("Subject: "))
		    subjectLine = true;
            	if (s.startsWith("Content-Type: "))
            	{
	            toPartProcessed = true;
	            out.println("To: " + to);
	            if (cc != null && cc.length() > 0)
	                out.println("cc: " + cc);
	            if (bcc != null && bcc.length() > 0)
	                out.println("bcc: " + bcc);
            	    while (s.startsWith("Content") || !isLetter(s.charAt(0)))
            	    {
            	    	++i;
            	    	if (i < addLines.size())
	            	    s = (String)addLines.get(i);
	            	else
	            	    break;
            	    }
            	    --i;
            	    s = "Content-Type: ";
		    if (withAttatchments && attachmentsFound.size() > 0)
			s += "multipart/mixed;\n  boundary=\"--" + boundary + "\"";
		    else
			s += guessedMime;
		    contentTypeLine = true;
            	}
                out.println(s);
            }
            if (!toPartProcessed)
            {
            	out.println("To: " + to);
            	if (cc != null && cc.length() > 0)
                    out.println("cc: " + cc);;
            	if (bcc != null && bcc.length() > 0)
                    out.println("bcc: " + bcc);
            }
	    if (!fromLine)
		out.println("From: " + from);
	    if (!dateLine)
		out.println("Date: " + date.toAscTimeInDateField());
	    if (!subjectLine)
		out.println("Subject: " + subject);
	    if (!contentTypeLine && withAttatchments && attachmentsFound.size() > 0)
		out.println("Content-Type: multipart/mixed;\n  boundary=\"--" + boundary + "\"");
            out.println();
    	    if (withAttatchments && attachmentsFound.size() > 0)
    	    {
    	    	out.println("----" + boundary);
    	    	out.println("Content-Type: " + guessedMime);
    	    }
	    for (i = 0; i < body.size(); i++)
	    {
		String s = (String)body.get(i);
		if (s.startsWith("From ") && s.length() > 5)
		{
		    int at = s.indexOf('@', 5);
		    int space = s.indexOf(' ', 5);
		    if (at > 0 && (space < 0 || at < space))
			s = ">" + s; // 'From email address' is terminating line!
		}
	    	out.println(s);
	    }
	    if (withAttatchments && attachmentsFound.size() > 0)
	    {
	    	for (i = 0; i < attachmentsFound.size(); i++)
	    	{
                    try
                    {
                        String fileName = (String)attachmentsFound.get(i);
                        File f = new File(fileName);
                        FileInputStream stream = new FileInputStream(f);
                        int length = (int)f.length();
                        if (length > maxFileSize)
                        {
                            System.out.println(); 
                            System.out.println("Attachment too long: " + fileName + ". Size: " + length + ", max file size: " + maxFileSize + ". Change maxFileSize variable to set new limit.");
                            stream.close();
                            continue;
                        }
                        byte [] buffer = new byte[length];
                        int read = stream.read(buffer);
                        if (read != length)
                            throw new Exception(fileName + ": " + read + " bytes read, " + length + " bytes expected");
                        out.println("\n----" + boundary);
                        String name = nameFromFullName(fileName);
                        out.println("Content-Type: " + contentTypeFromName(name) + "; name=\"" + name + "\"");
                        out.println("Content-Transfer-Encoding: base64");
                        out.println("Content-Disposition: attachment; FileName=\"" + name + "\"");
                        out.println();
                        encodeFile(buffer, out);
                    }
                    catch (Exception e)
                    {
                        e.printStackTrace();
                    }
	        }
    	        out.println("\n----" + boundary + "--");
	    }
            out.println();
        }
    }

    void encodeFile(byte [] buffer, PrintWriter out)
    {
	try
	{
	    StringBuffer temp = Base64.encode(buffer);
	    StringBuffer result = new StringBuffer(77);
	    int pos = -1, x = 0, len = temp.length();
	    while (x < len)
	    {
		char c = temp.charAt(x++);
		if (++pos > 76)
		{
		    pos = 0;
		    out.println(result);
		    result.setLength(0);
		}
		result.append(c);
	    }
	    out.println(result);
	}
	catch (Exception e)
	{
	    e.printStackTrace();
	}
    }
    
    private String nameFromFullName(String fullName)
    {
	int i = fullName.lastIndexOf('\\');
	if (i < 0)
	    return fullName;
	return fullName.substring(i+1);
    }

    private String contentTypeFromName(String name)
    {
	int i = name.lastIndexOf('.');
	if (i > 0)
	{
	    String ext = name.substring(i+1).toLowerCase();
	    if (ext.equals("gif"))
		return "image/gif";
	    else if (ext.equals("jpg") || ext.equals("jpeg"))
		return "image/jpeg";
	    else if (ext.equals("htm") || ext.equals("html"))
		return "text/html";
	    else if (ext.equals("txt"))
		return "text/plain";
	    else if (ext.equals("zip"))
		return "application/octet-stream";
	    else if (ext.equals("pdf"))
		return "application/pdf";
	    else if (ext.equals("exe"))
		return "application/x-msdownload";
	    else if (ext.equals("wav"))
		return "audio/wav";
	    return "application/octet-stream";
	}
	return "application/octet-stream";
    }
    
    private boolean isLetter(char c)
    {
        return (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z');
    }

    private boolean isBoundary(char c)
    {
        return isLetter(c) || c == '.' || c == ' ' || (c >= '0' && c <= '9');
    }

    private Mail readMail(int mailSize, int mwVersion) throws Exception
    {
//System.out.println("Size: " + mailSize);
        Mail mail = new Mail();
        int size = 4;
        mail.unknownInt1 = readInt();

        mail.subject = in.readLn();
        size += mail.subject.length()+2;

        mail.from1 = in.readLn();
        size += mail.from1.length()+2;

        mail.from2 = in.readLn();
        size += mail.from2.length()+2;

        mail.from3 = in.readLn();
        size += mail.from3.length()+2;

        mail.unknown1 = in.readLn();
        size += mail.unknown1.length()+2;

        mail.to = in.readLn();
        size += mail.to.length()+2;

        mail.cc = in.readLn();
        size += mail.cc.length()+2;

        mail.bcc = in.readLn();
        size += mail.bcc.length()+2;

	if (mwVersion == 0)
	{
            mail.id = in.readLn();
            size += mail.id.length()+2;
	}
	else if (mwVersion == 1)
	{
	    String s = in.readLn();
            size += s.length()+2;
	}

        mail.account = in.readLn();
        size += mail.account.length()+2;

	if (mwVersion == 0)
	{
            mail.mime = in.readLn();
            size += mail.mime.length()+2;

            mail.unknown3 = in.readLn();
            size += mail.unknown3.length()+2;

            mail.unknown4 = in.readLn();
            size += mail.unknown4.length()+2;

            mail.unknown5 = in.readLn();
            size += mail.unknown5.length()+2;

            mail.unknown6 = in.readLn();
            size += mail.unknown6.length()+2;
	}
	else
	    for (int i = 0; i < 8; i++)
	    {
             	String s = in.readLn();
            	size += s.length()+2;
	    }

        long l = readLong();
        mail.date = getDate(l);
        size += 8;
        
	int endOfBody = 0; // 3.61 only
	if (mwVersion == 1)
	{
	    readInt();
	    readInt();
	    size += 8;
	    endOfBody = readInt()+size+4;
	    size += 4;
	}
        mail.terminator = in.readLn();
        size += mail.terminator.length()+2;
        mail.boundary = "";
        for (int i = 0; i < mail.terminator.length(); i++)
            if (isBoundary(mail.terminator.charAt(i)))
            	mail.boundary += mail.terminator.charAt(i);

	if (mwVersion == 0)
            while(size < mailSize)
            {
            	String s = in.readLn();
            	size += s.length()+2;
            	if (s.equals(mail.terminator))
                    break;
            	mail.attachments.add(s);
            	File file = new File(s);
            	if (file.exists()/* && file.isFile()*/)
                    mail.attachmentsFound.add(s);
else {System.out.println(); System.out.println("Attatchment not found: " + s);}
            }

        while(size < mailSize)
        {
            String s = in.readLn();
            size += s.length()+2;
            if (s.equals(mail.terminator))
                break;
            mail.addLines.add(s);
        }

        mail.guessedMime = "text/plain; charset=\"iso-8859-1\"";
        mail.body = new Vector();
        while(size < mailSize)
        {
            String s = in.readLn();
	    if (s.toLowerCase().indexOf("<html>") >= 0)
        	mail.guessedMime = "text/html; charset=\"iso-8859-1\"";
            size += s.length()+2;
            if (mwVersion == 0 && s.equals(mail.terminator))
                break;
            mail.body.add(s);
            if (mwVersion == 1 && size >= endOfBody)
		break;
        }

	if (mwVersion == 1)
            while(true)
            {
//System.out.println("size: " + size + ", MailSize: " + mailSize);
		if (size >= mailSize)
		    break;
            	String s = in.readLn();
            	size += s.length()+2;
            	mail.attachments.add(s);
            	File file = new File(s);
            	if (file.exists())
                    mail.attachmentsFound.add(s);
            }

        pos += size;
        return mail;
    }

    private int readInt() throws Exception
    {
        int result = 0;
        result = in.read();
        int i = in.read();
        result += i*256;
        i = in.read();
        result += i*256*256;
        i = in.read();
        result += i*256*256*256;
        return result;
    }

    private long readLong() throws Exception
    {
        long result = readInt();
        if (result < 0)
            result += 0x100000000l;
        int i = readInt();
        result += 0x100000000l * i;
        return result;
    }

    private Date getDate(long l)
    {
        double d = Double.longBitsToDouble(l);
        double days = Math.floor(d);
        double fraction = d-days;
        fraction *= 24.0;
        double hour = Math.floor(fraction);
        fraction -= hour;
        fraction *= 60.0;
        double minute = Math.floor(fraction);
        fraction -= minute;
        fraction *= 60.0;
        double second = Math.floor(fraction);
        fraction -= second;
        l = (long)days-25569;
        l *= 24;
        l += (long)hour;
        l *= 60;
        l += (long)minute;
        l *= 60;
        l += (long)second;
        l *= 1000;
        l += Math.round(fraction * 1000.0);
        Date result = new Date();
        result.setTime(l);
        return result;
    }

/*    
    private byte readByte() throws Exception
    {
        int result = 0;
        result = in.read();
        return (byte)result;
    }
*/    
    
    public static void main(String [] args)
    {
    	boolean encodeBinaryAttatchments = true;
	int mwVersion = 0; 	// 0: 3.55
				// 1: 3.61
	File currentDir = new File(".");
	File [] files = currentDir.listFiles();
	int filesFound = 0;
	System.out.println("Seeking mail warrior 3.55 mail folders");
	for (int i = 0; i < files.length; i++)
            try
            {
        	Qdb qdb = new Qdb();
		if (files[i].isFile() && files[i].getName().toLowerCase().endsWith(".mwm"))
		{
		    filesFound++;
                    qdb.main(files[i], encodeBinaryAttatchments, mwVersion);
		}
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
	if (filesFound == 0)
	{
	    System.out.println("None found.");
	    System.out.println("Seeking mail warrior 3.61 mail folders");
	    mwVersion = 1;
	    for (int i = 0; i < files.length; i++)
            	try
            	{
		    if (files[i].isDirectory())
		    {
			File [] files2 = files[i].listFiles();
        		Qdb qdb = new Qdb();
			for (int j = 0; j < files2.length; j++)
			    if (files2[j].isFile() && files2[j].getName().toLowerCase().endsWith(".kmx"))
                            {
                                filesFound++;
                    	    	qdb.main(files2[j], encodeBinaryAttatchments, mwVersion);
                            }
		    }
            	}
            	catch (Exception e)
            	{
                    e.printStackTrace();
            	}
                if (filesFound == 0)
                    System.out.println("None found.");
	}
    }   
}


//*********************************************************************
//* Base64 - a simple base64 encoder and decoder.
//*
//*     Copyright (c) 1999, Bob Withers - bwit@pobox.com
//*
//* This code may be freely used for any purpose, either personal
//* or commercial, provided the authors copyright notice remains
//* intact.
//*********************************************************************


class Base64
{
    public static StringBuffer encode(byte[] data)
    {
        int c;
        int len = data.length;
        StringBuffer ret = new StringBuffer(((len / 3) + 1) * 4);
        for (int i = 0; i < len; ++i)
        {
            c = (data[i] >> 2) & 0x3f;
            ret.append(cvt.charAt(c));
            c = (data[i] << 4) & 0x3f;
            if (++i < len)
                c |= (data[i] >> 4) & 0x0f;

            ret.append(cvt.charAt(c));
            if (i < len)
            {
                c = (data[i] << 2) & 0x3f;
                if (++i < len)
                    c |= (data[i] >> 6) & 0x03;

                ret.append(cvt.charAt(c));
            }
            else
            {
                ++i;
                ret.append((char) fillchar);
            }

            if (i < len)
            {
                c = data[i] & 0x3f;
                ret.append(cvt.charAt(c));
            }
            else
            {
                ret.append((char) fillchar);
            }
        }

        return(ret);
    }

    public static byte[] decode(byte[] data)
    {
        int c;
        int c1;
        int len = data.length;
        StringBuffer ret = new StringBuffer((len * 3) / 4);
        for (int i = 0; i < len; ++i)
        {
            c = cvt.indexOf(data[i]);
            ++i;
            c1 = cvt.indexOf(data[i]);
            c = ((c << 2) | ((c1 >> 4) & 0x3));
            ret.append((char) c);
            if (++i < len)
            {
                c = data[i];
                if (fillchar == c)
                    break;

                c = cvt.indexOf((char) c);
                c1 = ((c1 << 4) & 0xf0) | ((c >> 2) & 0xf);
                ret.append((char) c1);
            }

            if (++i < len)
            {
                c1 = data[i];
                if (fillchar == c1)
                    break;

                c1 = cvt.indexOf((char) c1);
                c = ((c << 6) & 0xc0) | c1;
                ret.append((char) c);
            }
        }

        return(getBinaryBytes(ret.toString()));
    }

    private static String getString(byte[] arr)
    {
        StringBuffer buf = new StringBuffer();
        for (int i = 0; i < arr.length; ++i)
            buf.append((char) arr[i]);

        return(buf.toString());
    }

    private static byte[] getBinaryBytes(String str)
    {
        byte[] b = new byte[str.length()];
        for (int i = 0; i < b.length; ++i)
            b[i] = (byte) str.charAt(i);

        return(b);
    }

    private static final int    fillchar = '=';

                                    // 00000000001111111111222222
                                    // 01234567890123456789012345
    private static final String cvt = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

                                    // 22223333333333444444444455
                                    // 67890123456789012345678901
                                    + "abcdefghijklmnopqrstuvwxyz"

                                    // 555555556666
                                    // 234567890123
                                    + "0123456789+/";
}
